# Name:           line2cyl
# Author :        Didier Bur
# Description:    Create cylinders with selected edges
# Usage:          select lines, enter diameter and precision
#
# Type:           Tool
# Date:           2005,13,01
# Revised:        2005,17,01: add the arcs and circles followme's

module SU::Powertools

	### MENU & TOOLBARS ### --------------------------------------------------

	unless file_loaded?( __FILE__ )      
		cmd_Lines2tubes = UI::Command.new('线转圆柱') { Lines2tubes.line2cyl }
		cmd_Lines2tubes.large_icon = 'Images/Lines2tubes_24.png'
		cmd_Lines2tubes.small_icon = 'Images/Lines2tubes_16.png'
		cmd_Lines2tubes.status_bar_text = '线转圆柱：焊接的线会连续，不焊接的线会打断。'
		cmd_Lines2tubes.tooltip = '线转圆柱'

		# Menu
		@menu.add_item( cmd_Lines2tubes )
		#@menu.add_separator

		# Toolbar
		@toolbar.add_item( cmd_Lines2tubes )
		#@toolbar.add_separator
	end # UI
	file_loaded( __FILE__ )
  
  
	### METHODS ### ----------------------------------------------------------
	class Lines2tubes

		def self.line2cyl
		   model = Sketchup.active_model
		   m_ents = model.active_entities
		   materials=model.materials
		   ss = model.selection
		   ignored = 0
		   ss_edges=[]
		   ss_alone = []
		   ss_curves = []
		   ss_connected = []
		   trunk=[]
		   $l2c_diam=10.cm if not $l2c_diam
		   $l2c_segs=10 if not $l2c_segs
		   $l2c_opt="Yes" if not $l2c_opt
		   $l2c_mat="Default" if not $l2c_mat
		   $l2c_grp1="Yes" if not $l2c_grp1
		   $l2c_grp2="No" if not $l2c_grp2

		   Sketchup.set_status_text("")

		   model.start_operation "Lines -> Cylinders"
		   # Selection error checking
		   if ss.empty?
			  UI.messagebox("No selection.")
			  return nil
		   end
		   # Dialog
		   list2=materials.collect { |m| m.name }
		   if list2.length==0
			  list2=["Default"]
		   end
		   drops=[["Yes","No"].join("|"),["Yes","No"].join("|"),["Yes","No"].join("|"),list2.join("|")]
		   prompts = ["跟随线条: ","创建群组: ","单独成组: ","材质选择: ","圆柱直径: ", "圆柱段数:  "]
		   values = [$l2c_opt,$l2c_grp1,$l2c_grp2,$l2c_mat,$l2c_diam,$l2c_segs]
		   results = inputbox prompts, values,drops, "Parameters"
		   return if not results

		   $l2c_opt = results[0]
		   $l2c_grp1 = results[1]
		   $l2c_grp2 = results[2]
		   $l2c_mat = results[3]
		   $l2c_diam = results[4]
		   $l2c_segs = results[5]

		   # Create group if needed
		   if $l2c_grp1=="Yes"
			  grp=m_ents.add_group
			  ents=grp.entities
		   else
			  ents=m_ents
		   end
		   # Dispatch selection
		   pbcount=0
		   #Sketchup.set_status_text("Parsing selection...")
		   ss.each do |x|
			  ss_edges.push(x) if x.typename=="Edge"
			  pbcount+=1
		   end
		   ignored=ss.length-ss_edges.length

		   if $l2c_opt=="Yes"
			  #Sketchup.set_status_text("Dispatching edges...")
			  pbcount=0
			  ss_edges.each do |e|
				 if e.curve
					ss_curves.push(e.curve)
				 else
					ss_alone.push(e)
				 end
				 #if e.all_connected.length>1 and not e.curve
				 #ss_connected.push(e.all_connected)
				 #end
				 #if e.all_connected.length==1
				 #ss_alone.push(e)
				 #end
				 pbcount+=1
			  end
			  ss_curves.uniq!
			  #ss_connected.uniq!
			  ss_alone.uniq!

			  #Sketchup.set_status_text("Generating cylinders...")
			  pbcount=0
			  # Pushpull the standalone edges in their own group
			  ss_alone.each do |e|
				 vec = e.line[1]
				 tube_group=ents.add_group
				 tube_ents=tube_group.entities
				 circle = tube_ents.add_circle(e.vertices[0], vec, ($l2c_diam / 2.0), $l2c_segs)
				 if circle
					base = tube_ents.add_face(circle)
					base.material=$l2c_mat if $l2c_mat!="Default"
					if base
					   base.reverse! if base.normal != vec
					   base.pushpull e.length
					end
					# Store tube group in a trunk
					trunk.push(tube_group)
				 end
				 pbcount+=1
			  end

			  # Followme's on curves
			  ss_curves.each do |c|
				 edge1 = c.first_edge
				 vec = edge1.line[1]
				 ext1 = edge1.start.position
				 all_edges = c.edges
				 tube_group=ents.add_group
				 tube_ents=tube_group.entities
				 circle = tube_ents.add_circle(ext1, vec, ($l2c_diam / 2.0), $l2c_segs)
				 if circle
					base = tube_ents.add_face(circle)
					base.material=$l2c_mat if $l2c_mat!="Default"
					if base
					   base.reverse!
					   base.followme all_edges
					end
					# Store tube group in a trunk
					trunk.push(tube_group)
				 end
				 pbcount+=1
			  end

			  ss_connected.each do |conn_e|
				 # find first edge
				 ext_edges=[]
				 conn_e.each { |e| ext_edges.push(e) if e.start.edges.length==1;
				 ext_edges.push(e) if e.end.edges.length==1 }

				 vec = ext_edges[0].line[1]
				 tube_group=ents.add_group
				 tube_ents=tube_group.entities
				 circle = tube_ents.add_circle(ext1, vec, ($l2c_diam / 2.0), $l2c_segs)
				 if circle
					base = tube_ents.add_face(circle)
					base.material=$l2c_mat if $l2c_mat!="Default"
					if base
					   #base.reverse!
					   base.followme conn_e
					end
					# Store tube group in a trunk
					trunk.push(tube_group)
				 end
			  end
			  Sketchup.set_status_text("")
		   end

		   if $l2c_opt=="No"
			  pbcount=0
			  # Pushpull the standalone edges in their own group
			  ss_edges.each do |e|
				 vec = e.line[1]
				 tube_group=ents.add_group
				 tube_ents=tube_group.entities
				 circle = tube_ents.add_circle(e.start.position, vec, ($l2c_diam / 2.0), $l2c_segs)
				 if circle
					base = tube_ents.add_face(circle)
					if base
					   base.material=$l2c_mat if $l2c_mat!="Default"
					   base.reverse! if base.normal != vec
					   base.pushpull e.length
					end
					# Store tube group in a trunk
					trunk.push(tube_group)
				 end
				 pbcount+=1
			  end
		   end

		   Sketchup.set_status_text("生成柱体，请等待...")
		   # Explode all tube groups
		   if $l2c_grp2=="No"
			  trunk.each { |tube| tube.explode if tube.valid? }
		   end
		   UI.messagebox(ss_edges.length.to_s+" 柱体生成,\n"+ignored.to_s+" 个实体不是线条被忽略.")
		   Sketchup.set_status_text("")
		   model.commit_operation
		end #of def

	end
  
end # module