module SU::Powertools
	# Copyright 2004,2005 by Rick Wilson 

	# Permission to use, copy, modify, and distribute this software for 
	# any purpose and without fee is hereby granted, provided that the above
	# copyright notice appear in all copies.

	# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
	# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
	# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

	# Name :          Weld 2.3
	# Description :   weld edges into curve ("polyline")
	# Author :        RickW
	# Usage :         select edges, then run the script (Plugins>Weld)
	#			Edges will be joined into a curve ("polyline")
	#			Script will prompt to find faces
	# Date :          4.Aug.2004
	# Type :		tool
	# History:
	#			3.0 (4.Aug.2005) -  Special Anniversary Edition! :)
	#						  now welds to existing curves, adds option
	#						  to find faces
	#			2.2 (1.Aug.2005) -  used group/explode to fix most problems
	#						  still won't weld to non-arc curves
	#			2.1 (27.Jul.2005) - fixed the fix that caused strange non-welding
	#						  issues.
	#			2.0 (25.Jul.2005) - fixed bug where unselected edges were welded,
	#						  now retains existing faces.
	#			1.2 (12.May.2005) - fixed edge bug that occasionally resulted
	#						  in strange edge locations and deletions
	#			1.1 (10.Sep.2004) - allows working in groups/components
	#			1.0 (4.Aug.2004) - first version
	### MENU & TOOLBARS ### --------------------------------------------------

	unless file_loaded?( __FILE__ )      
		cmd_weld = UI::Command.new('焊接线段') { Weld.join_edges }
		cmd_weld.large_icon = 'Images/weld_24.png'
		cmd_weld.small_icon = 'Images/weld_16.png'
		cmd_weld.status_bar_text = '选择两段或以上的线段进行焊接。'
		cmd_weld.tooltip = '焊接线段'

		# Menu
		@menu.add_item( cmd_weld )
		#@menu.add_separator

		# Toolbar
		@toolbar.add_item( cmd_weld )
		#@toolbar.add_separator
	end # UI
	file_loaded( __FILE__ )
  
  
	### METHODS ### ----------------------------------------------------------
	class Weld
		def self.join_edges
			model=Sketchup.active_model
			ents=model.active_entities
			sel=model.selection
			sl=sel.length
			verts=[]
			edges=[]
			newVerts=[]
			startEdge=startVert=nil

		#CHECK FOR A VALID SELECTION

		#	if sel.first.curve
		#		return UI.messagebox("Please select at least 2 non-joined edges") if sel.length == sel.first.curve.count_edges
		#	end

		#DELETE NON-EDGES, GET THE VERTICES

			sel.each {|item| edges.push(item) if item.typename=="Edge"}
			edges.each {|edge| verts.push(edge.vertices)}
			verts.flatten!
			return UI.messagebox("请至少选择两段线段！") if edges.length < 2

		#FIND AN END VERTEX

			vertsShort=[]
			vertsLong=[]
			verts.each do |v|
				if vertsLong.include?(v)
					vertsShort.push(v)
				else
					vertsLong.push(v)
				end
			end
			if (startVert=(vertsLong-vertsShort).first)==nil
				startVert=vertsLong.first
				closed=true
				startEdge = startVert.edges.first
			else
				closed=false
				startEdge = (edges & startVert.edges).first
			end
			#sel.clear


		#SORT VERTICES, LIMITING TO THOSE IN THE SELECTION SET

			if startVert==startEdge.start
				newVerts=[startVert]
				counter=0
				while newVerts.length < verts.length
					edges.each do |edge|
						if edge.end==newVerts.last
							newVerts.push(edge.start)
						elsif edge.start==newVerts.last
							newVerts.push(edge.end)
						end
					end
					counter+=1
					if counter > verts.length
						return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
						newVerts.reverse!
						reversed=true
					end
				end
			else
				newVerts=[startVert]
				counter=0
				while newVerts.length < verts.length
					edges.each do |edge|
						if edge.end==newVerts.last
							newVerts.push(edge.start)
						elsif edge.start==newVerts.last
							newVerts.push(edge.end)
						end
					end
					counter+=1
					if counter > verts.length
						return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
						newVerts.reverse!
						reversed=true
					end
				end
			end
			newVerts.uniq!
			newVerts.reverse! if reversed


		#CONVERT VERTICES TO POINT3Ds

			newVerts.collect!{|x| x.position}
			if closed
				newVerts.push(newVerts[0])
			end

		#CREATE THE CURVE

			model.start_operation "焊接线段"
			edges.each {|item| item.explode_curve if item.curve and newVerts.include?(item.start.position)}
			#sel.each {|e| e.erase!}
			gp1e=(gp1=ents.add_group).entities
			curveEdges=gp1e.add_curve(newVerts)
			gp1.explode
			#ents[-1].find_faces if closed and ents[-1].typename=="Edge"
			model.commit_operation
			sel.clear
		end

	end
  
end # module